#page_info {
	background-color: rgba(16, 16, 16, 0.70);
	color: white;
	padding: 5px;
	padding-left: 10px;
	padding-right: 10px;
}

#page_info .button-left {
	border-right-width: 0px;
}

#page_info > box > label {
	color: #919693;
}

#page_info {
	color: gray;
}

#page_info button {
	color: white;
	background-image: none;
	background-color: #373737;
	border: 1px solid #202121;
}

#page_info button:disabled {
	color: gray;
}
