CAIRO = {
    'debian': 'python3-gi-cairo',
    'fedora': 'python3-pycairo',
    'gentoo': 'dev-python/pycairo',  # Python 3 ?
    'linuxmint': 'python-gi-cairo',  # Python 3 ?
    'raspbian': 'python3-gi-cairo',
    'suse': 'python-cairo',  # Python 3 ?
    'ubuntu': 'python3-gi-cairo',
}

GDK = {
    'debian': 'gir1.2-gdkpixbuf-2.0',
    'linuxmint': 'gir1.2-gdkpixbuf-2.0',
    'raspbian': 'gir1.2-gdkpixbuf-2.0',
    'ubuntu': 'gir1.2-gdkpixbuf-2.0',
}

GI = {
    'debian': 'python3-gi',
    'fedora': 'python3-gobject-base',
    'gentoo': 'dev-python/pygobject',  # Python 3 ?
    'linuxmint': 'python3-gi',
    'raspbian': 'python3-gi',
    'suse': 'python-gobject',  # Python 3 ?
    'ubuntu': 'python3-gi',
}

GLIB = {
    'debian': 'gir1.2-glib-2.0',
    'linuxmint': 'gir1.2-glib-2.0',
    'raspbian': 'gir1.2-glib-2.0',
    'ubuntu': 'gir1.2-glib-2.0',
}

PANGO = {
    'debian': 'gir1.2-pango-1.0',
    'linuxmint': 'gir1.2-pango-1.0',
    'raspbian': 'gir1.2-glib-2.0',
    'ubuntu': 'gir1.2-pango-1.0',
}

POPPLER = {
    'debian': 'gir1.2-poppler-0.18',
    'fedora': 'poppler-glib',
    'gentoo': 'app-text/poppler',
    'linuxmint': 'gir1.2-poppler-0.18',
    'raspbian': 'gir1.2-poppler-0.18',
    'suse': 'typelib-1_0-Poppler-0_18',
    'ubuntu': 'gir1.2-poppler-0.18',
}
